/*
 * Copyright (C) 2013 Canonical, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import QtQuick 2.0
import "../Components"
import Ubuntu.Components 0.1
import Ubuntu.Gestures 0.1
import "../Applications/applications.js" as ApplicationsModel

Item {
    id: launcher

    property var favourites: ["/usr/share/applications/phone-app.desktop",
                              "/usr/share/applications/camera-app.desktop",
                              "/usr/share/applications/gallery-app.desktop",
                              "/usr/share/applications/facebook-webapp.desktop",
                              "/usr/share/applications/webbrowser-app.desktop",
                              "/usr/share/applications/twitter-webapp.desktop",
                              "/usr/share/applications/gmail-webapp.desktop",
                              "/usr/share/applications/ubuntu-weather-app.desktop",
                              "/usr/share/applications/notes-app.desktop",
                              "/usr/share/applications/ubuntu-calendar-app.desktop"]
    property ListModel model: ListModel {}
    property ListModel applicationsRunning: ListModel {}
    property int lastFavouriteIndex: -1

    property bool available: true // can be used to disable all interactions
    property bool applicationFocused: false

    property bool shown: launcher.state != ""
    property int progress
    progress: {
        if (launcher.state == "spreadMoving")
            return dragArea.distance
        else if (launcher.state == "spread")
            return launcher.width
        else
            return 0
    }

    // Thresholds and dimensions
    property int shortcutsHintWidth: units.gu(3)
    property int shortcutsWidth: units.gu(6)
    property int shortcutsThreshold: units.dp(2)
    property int shortcutsCloseThreshold: units.gu(1)
    property int spreadHintThreshold: shortcutsWidth
    property int spreadOutThreshold: units.gu(18)

    // For how long (in milliseconds) the launcher will stay up without being
    // interacted with, before it finally hides itself.
    property int inactivityTimeout: 8000

    property string iconPath: "."

    property int launcherProgressValue
    launcherProgressValue: {
        if (launcher.state == ""
                || launcher.state == "hint"
                || launcher.state == "moving"
                || launcher.state == "shortcuts") {
            return 0
        } else if (launcher.state == "spreadHint") {
            return units.gu(17)
        } else if (launcher.state == "spreadMoving") {
            return units.gu(22)
        } else if (launcher.state == "spread") {
            return units.gu(28)
        }
    }

    property bool teasing: false

    signal launcherApplicationSelected(string name)
    signal dashItemSelected(int itemIndex)

    signal dash()

    onTeasingChanged: {
        if (teasing) {
            hintTimer.restart();
        }
    }

    Timer {
        id: hintTimer
        interval: 200
    }

    // internal properties hidden from public api
    QtObject {
        id: priv

        property bool shortcutAnimationOngoing
        property string previousStateHelp
        property string previousState: ""
        property string teasing: launcher.teasing || hintTimer.running
    }

    onStateChanged: {
        priv.previousState = priv.previousStateHelp
        priv.previousStateHelp = state
        if (launcher.state != "shorcuts") {
            dismissTimer.stop()
        }

        if (launcher.state == "spread") {
            dragArea.visible = false
            hideTimer.restart()
            dragArea.visible = true
        }
        //console.log("Launcher state: " + state)
    }

    Component.onCompleted: {
        __populateWithFavouriteApplications();
    }

    function __populateWithFavouriteApplications() {
        for (var i=0; i<launcher.favourites.length; i++) {
            var desktopFile = launcher.favourites[i];
            var application = ApplicationsModel.__availableApplications[desktopFile];
            if (application) {
                application["desktopFile"] = desktopFile;
                launcher.model.append(application);
            }
        }
    }

    function hide() {
        launcher.state = ""
    }

    Rectangle {
        id: backgroundShade
        anchors.fill: parent
        color: "black"
        opacity: {
            if (launcher.state == "" || launcher.state == "hint" || launcher.state == "moving") {
                return 0
            } else return 0.4
        }
        visible: opacity != 0

        Behavior on opacity { NumberAnimation { duration: 200; easing.type: Easing.OutQuad} }

        MouseArea {
            anchors.fill: parent
            onClicked: {
                launcher.state = ""
            }
            visible: launcher.state != "spread"
        }
    }

    LauncherShortcuts {
        id: shortcuts
        objectName: "shortcuts"
        width: shortcutsWidth
        anchors {
            top: parent.top
            bottom: parent.bottom
        }
        x: -width
        model: launcher.model
        onApplicationSelected: {
            launcher.state = ""
            launcherApplicationSelected(desktopFile)
        }
        onDashItemSelected: {
            launcher.state = ""
            launcher.dashItemSelected(index)
        }
        iconPath: launcher.iconPath

        Behavior on x { NumberAnimation {duration: launcher.state == "hint" ? 70 : launcher.state == "moving" ? 0 : 300; easing.type: Easing.OutCubic } }
    }

    DirectionalDragArea {
        id: dragArea
        enabled: launcher.state != "shortcuts" || dragging

        direction: DirectionalDragArea.Rightwards

        // values to be tweaked and later removed from here and set in stone as defaults
        // once we are confident it's all good.
        maxDeviation: units.gu(1.5)
        wideningAngle: 10
        distanceThreshold: units.gu(4)

        visible: launcher.available
        width: {
            if (launcher.state == "") {
                return shortcutsThreshold
            } else {
                return shortcutsWidth
            }
        }
        height: launcher.height

        onDraggingChanged: {
            if (dragging)
                onDragStarted()
            else
                onDragEnded()
        }

        function onDragStarted() {
            if (launcher.state == "") launcher.state = "hint"
            dismissTimer.stop()
        }

        onDistanceChanged: {
            if (launcher.state == "shortcuts" && distance < shortcutsCloseThreshold) {
                launcher.state = ""
            } else if (launcher.state == "hint" && distance > 0 ) {
                launcher.state = "moving"
            } else if (distance > shortcutsThreshold && distance < spreadHintThreshold) {
                launcher.state = "shortcuts"
            } else if (distance >= spreadHintThreshold && distance < shortcutsWidth) {
                if (launcher.applicationFocused){
                    launcher.state = "spreadHint"
                }
            } else if (distance >= shortcutsWidth && distance < spreadOutThreshold) {
                if (launcher.applicationFocused) {
                    launcher.state = "spreadMoving"
                } else {
                    dismissTimer.restart()
                }
            } else if (distance >= spreadOutThreshold && launcher.applicationFocused) {
                launcher.state = "spread"
            }
        }

        function onDragEnded() {
            var currentApplicationIndex
            if (launcher.state == "shortcuts" && !priv.shortcutAnimationOngoing) {
                // launch app
                currentApplicationIndex = shortcuts.selectCurrentApplication()
                if (currentApplicationIndex === 0) {
                    launcher.dashItemSelected(0)
                    launcher.state = ""
                } else if (currentApplicationIndex === -1) {
                    launcher.state = "shortcuts"
                }
            }

            if (launcher.state == "hint" || launcher.state == "moving") {
                launcher.state = ""
            } else if (launcher.state == "spreadMoving" || launcher.state == "spreadHint") {
                launcher.state = "shortcuts"
            }

            shortcuts.stopScroll()

            if (launcher.state == "shortcuts") {
                dismissTimer.restart()
            }

        }
    }

    Timer {
        id: dismissTimer
        interval: launcher.inactivityTimeout;
        running: false;
        repeat: false
        onTriggered: {
            launcher.state = ""
        }
    }

    Timer {
        id: hideTimer
        interval: 50;
        running: false;
        repeat: false
        onTriggered: {
            launcher.state = ""
        }
    }

    states: [
        State {
            name: "hint"
            when: priv.teasing
            PropertyChanges {
                target: shortcuts
                x: -shortcuts.width + launcher.shortcutsHintWidth
            }
        },
        State {
            name: "moving"
            PropertyChanges {
                target: shortcuts
                x: -shortcuts.width + dragArea.distance + launcher.shortcutsHintWidth
            }
        },
        State {
            name: "shortcuts"
            PropertyChanges {
                target: shortcuts
                x: 0
            }
        },
        State {
            name: "spreadHint"
            extend: "shortcuts"
        },
        State {
            name: "spreadMoving"
            PropertyChanges {
                target: launcher
                launcherProgressValue: dragArea.distance
            }
            PropertyChanges {
                target: shortcuts
                x: 0
            }
        },
        State {
            name: "spread"
            extend: "shortcuts"
        }

    ]

    transitions: [
        Transition {
            to: ""
            SequentialAnimation {
                PropertyAnimation {
                    target: shortcuts
                    properties: "x"
                    duration: 400
                    easing.type: Easing.OutCubic
                }
                ScriptAction {
                    script: {
                        shortcuts.resetScrollPosition()
                    }
                }
            }
        },
        Transition {
            to: "hint"
            PropertyAnimation {
                target: shortcuts
                properties: "x"
                duration: 400
                easing.type: Easing.OutCubic
            }
        },
        Transition {
            to: "shortcuts"
            SequentialAnimation {
                ScriptAction {
                    script: {
                        if (priv.previousState == "moving") {
                            priv.shortcutAnimationOngoing = true
                        }
                    }
                }
                ParallelAnimation {
                    SequentialAnimation {
                        PropertyAnimation {
                            target: shortcuts
                            properties: "x"
                            duration: 300
                            easing.type: Easing.OutCubic
                        }
                        ScriptAction {
                            script: {
                                priv.shortcutAnimationOngoing = false
                            }
                        }
                    }
                }
            }
        },
        Transition {
            to: "spreadHint"
            PropertyAnimation {
                target: shortcuts
                properties: "x"
                duration: 300
                easing.type: Easing.OutCubic
            }
        },
        Transition {
            to: "spreadMoving"
            ScriptAction {
                script: {
                    priv.shortcutAnimationOngoing = false
                }
            }
        },
        Transition {
            to: "spread"
            ScriptAction {
                script: {
                    launcher.dash()
                }
            }
        }
    ]
}
